#' @title load_primary_netcdf
#' @description Load the primary netcdf file (`simu_results_0.nc`) in the given path
#'
#' @param path The path to the directory containing camp simulations netcdf results
#'
#' @return an object of class `ncdf4`
#'
load_primary_netcdf <- function(path) {
  nc_file <- paste0(path, "/simu_results_0.nc")
  return(nc_file)
}

#' @title load_secondary_netcdf
#' @description Load secondary netcdf files (any `.nc` file except `simu_results_0.nc`) in the given path
#' They are picked in a random order, and the number of loaded files is limited by `max_to_take`
#'
#' @param path The path to the directory containing camp simulations netcdf results
#' @param max_to_take (default = 112) maximum number of files to load
#'
#' @return a list of objects of class `ncdf4`
#'
load_secondary_netcdf <- function(path, max_to_take = 112) {
  file_list <- dir(
    path,
    pattern = glob2rx("simu_results_*.nc"),
    full.names = TRUE
  )

  # and randomly pick simulations except the first one ("simu_results_0")
  if (max_to_take > 1 && length(file_list) > 1) {
    file_list <- file_list[2:(min(length(file_list), max_to_take + 1))]
    ncfile_list <- file_list
  } else {
    ncfile_list <- NULL
  }


  return(ncfile_list)
}


#' @title get_gas_names
#' @description read the primary netcdf output file and outputs a data.table
#' matching `gas_species_dim` with `gas_species_name`
#'
#' @param path The path to the directory containing camp simulations netcdf results
#'
#' @return a data.table object with two columns: `gas_species_dim` `gas_species_name`
#'
get_gas_names <- function(path) {
  nc_file <- load_primary_netcdf(path)
  gas_names <- CustomReadNetCDF(nc_file, vars = "gas_species_name")
  return(gas_names)
}

#' @title list_gas_names
#' @description get the list of gas phase species
#'
#' @param path  The path to the directory containing camp simulations netcdf results
#'
#' @return a character vector of species names
#' @export
list_gas_names <- function(path) {
  gas_names <- get_gas_names(path)
  return(gas_names$gas_species_name)
}

#' @title get_aer_names
#' @description read the primary netcdf output file and outputs a data.table
#' matching `aer_species_dim` and `aer_phases_dim` with `aer_species_name`
#'
#' @param path The path to the directory containing camp simulations netcdf results
#'
#' @return a data.table object with three columns:  `aer_species_dim` `aer_phases_dim` `aerosol_species_name`
#' @export
get_aer_names <- function(path) {
  nc_file <- load_primary_netcdf(path)
  aer_names <- CustomReadNetCDF(nc_file, vars = "aerosol_species_name")
  return(aer_names)
}

#' @title list_gas_names
#' @description get the list of gas phase species
#'
#' @param path  The path to the directory containing camp simulations netcdf results
#'
#' @return a character vector of species names
#' @export
list_aer_names <- function(path) {
  aer_names <- get_aer_names(path)
  return(aer_names$aerosol_species_name)
}

#' @title get_phase_names
#' @description read the primary netcdf output file and outputs a data.table
#' matching `aero_phases_dim` with `aerosol_phase_name`
#'
#' @param path The path to the directory containing camp simulations netcdf results
#'
#' @return a data.table object with two columns:  `aero_phases_dim` `aerosol_phase_name`
#'
get_phase_names <- function(path) {
  nc_file <- load_primary_netcdf(path)
  phase_names <- CustomReadNetCDF(nc_file, vars = "aerosol_phase_name")
  return(phase_names)
}

#' @title list_phase_names
#' @description get the list of aerosol phases
#'
#' @param path  The path to the directory containing camp simulations netcdf results
#'
#' @return a character vector of phase names
#' @export
list_phase_names <- function(path) {
  phase_names <- get_phase_names(path)
  return(phase_names$aerosol_phase_name)
}

#' @title get_var_from_primary_cell
#'
#' @param path The path to the directory containing camp simulations netcdf results
#' @param varname The name of the variable to read
#'
#' @return a data.table
#' @export
get_var_from_primary_cell <- function(path, varname) {
  ncfile <- load_primary_netcdf(path)
  dt <- metR::ReadNetCDF(ncfile, vars = varname, key = TRUE)
  return(dt)
}

#' @title get_var_from_all_secondary_cells
#'
#' @param path The path to the directory containing camp simulations netcdf results
#' @param varname The name of the variable to read
#' @param max_to_take (default = 112) maximum number of files to load
#'
#' @return a data.table
#' @export
get_var_from_all_secondary_cells <- function(path, varname, max_to_take = 112) {
  ncfile_list <- load_secondary_netcdf(path, max_to_take - 1)
  if (!is.null(ncfile_list)) {
    dt_list <- lapply(
      ncfile_list, function(x) {
        metR::ReadNetCDF(x, vars = varname, key = TRUE)
      }
    )
    dt <- do.call("rbind", dt_list)
  } else {
    dt <- NULL
  }

  return(dt)
}

#' @title get_var_from_all_cells
#'
#' @param path The path to the directory containing camp simulations netcdf results
#' @param varname The name of the variable to read
#' @param max_to_take (default = 112) maximum number of files to load
#'
#' @return a data.table
#' @export
get_var_from_all_cells <- function(path, varname, max_to_take = 112) {
  primary_dt <- get_var_from_primary_cell(path, varname)
  result <- primary_dt
  if (max_to_take > 1) {
    secondary_dt <- get_var_from_all_secondary_cells(path, varname, max_to_take)
    if (!is.null(secondary_dt)) {
      result <- rbind(primary_dt, secondary_dt)
    }
  }

  return(result)
}

#' @title get_gas_species_concentrations
#' @description outputs the gas phase concentration of the selected species for all cells in given path
#'
#' @param path The path to the directory containing camp simulations netcdf results
#' @param species_list the list of gas phase species we want to get
#' @param max_to_take (default = 112) maximum number of files to load
#'
#' @return a data.table object
#'
#' @export
get_gas_species_concentrations <- function(path, species_list, max_to_take = 112) {
  gas_names <- get_gas_names(path)
  dt <- get_var_from_all_cells(path, "gas_species_concentrations", max_to_take)
  dt[
    gas_names,
    on = c("gas_species_dim"),
    species := gas_species_name
  ]

  return(dt[species %in% species_list])
}

#' @title get_aer_species_bulk_concentrations
#' @description outputs the aerosol phase concentration of the selected species for all cells in given path  (kg/m3)
#'
#' @param path The path to the directory containing camp simulations netcdf results
#' @param species_list the list of aerosol phase species we want to get
#' @param max_to_take (default = 112) maximum number of files to load
#'
#' @return a data.table object
#'
#' @export
get_aer_species_bulk_concentrations <- function(path, species_list, max_to_take = 112) {
  aer_names <- get_aer_names(path)
  dt <- get_var_from_all_cells(path, "aerosol_species_concentration", max_to_take)
  dt <- dt[
    aer_names,
    on = c("aero_phases_dim", "aero_species_dim")
  ]
  dt <- dt[aerosol_species_name %in% species_list, ]
  # for some reason this doesn't work with data.table
  # need to switch to tidyverse
  # dt <- dt[
  #   ,
  #   .(bulk_concentration_kg_m3 = sum(aerosol_species_concentration)),
  #   by = .(time, cell, aerosol_species_name)
  # ]
  dt <- dt %>%
    dplyr::group_by(time, cell, aerosol_species_name) %>%
    dplyr::summarise(bulk_concentration_kg_m3 = sum(aerosol_species_concentration))
  data.table::setDT(dt)
  dt
}

#' @title get_aer_phases_bulk_concentrations
#' @description outputs the aerosol phase concentration split by phase (kg/m3)
#'
#' @param path The path to the directory containing camp simulations netcdf results
#' @param max_to_take (default = 112) maximum number of files to load
#'
#' @return a data.table object
#'
#' @export
get_aer_phases_bulk_concentrations <- function(path, max_to_take = 112) {
  phase_names <- get_phase_names(path)
  dt <- get_var_from_all_cells(path, "aerosol_species_concentration", max_to_take)
  dt <- dt[
    phase_names,
    on = c("aero_phases_dim")
  ]
  # dt[
  #   ,
  #   .(bulk_concentration_kg_m3 = sum(aerosol_species_concentration)),
  #   by = c("time", "cell", "aerosol_phase_name")
  # ]
  dt <- dt %>%
    dplyr::group_by(time, cell, aerosol_phase_name) %>%
    dplyr::summarise(bulk_concentration_kg_m3 = sum(aerosol_species_concentration))
  data.table::setDT(dt)
  dt
}

#' @title get_aer_total_bulk_concentrations
#' @description outputs the total aerosol phase concentration (kg/m3)
#'
#' @param path The path to the directory containing camp simulations netcdf results
#' @param max_to_take (default = 112) maximum number of files to load
#'
#' @return a data.table object
#'
#' @export
get_aer_total_bulk_concentrations <- function(path, max_to_take = 112) {
  dt <- get_var_from_all_cells(path, "aerosol_species_concentration", max_to_take)

  # dt <- dt[
  #   ,
  #   .(bulk_concentration_kg_m3 = sum(aerosol_species_concentration)),
  #   by = c("time", "cell")
  # ]
  dt <- dt %>%
    dplyr::group_by(time, cell) %>%
    dplyr::summarise(bulk_concentration_kg_m3 = sum(aerosol_species_concentration))
  data.table::setDT(dt)
  dt
}
#' @title get_aer_secondary_bulk_concentrations
#' @description outputs the secondary aerosol phase concentration (kg/m3)
#'
#' @param path The path to the directory containing camp simulations netcdf results
#' @param primary_species a character vector of the species that are considered primary and should
#' be removed from the aerosol mass
#' @param max_to_take (default = 112) maximum number of files to load
#'
#' @return a data.table object
#'
#' @export
get_aer_secondary_bulk_concentrations <- function(path, primary_species, max_to_take = 112) {
  aer_names <- get_aer_names(path)
  dt <- get_var_from_all_cells(path, "aerosol_species_concentration", max_to_take)
  dt <- dt[
    aer_names,
    on = c("aero_phases_dim", "aero_species_dim")
  ]
  dt <- dt[!aerosol_species_name %in% primary_species, ]
  dt <- dt %>%
    dplyr::group_by(time, cell) %>%
    dplyr::summarise(bulk_concentration_kg_m3 = sum(aerosol_species_concentration))
  data.table::setDT(dt)
  dt
}
